# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
import avro
import cStringIO as StringIO
import httplib
import logging
import os
import sys
import tarfile
import tempfile
import threading
import threadpool
import time
import urllib
import urllib2
import urlparse

LOG = logging.getLogger(__name__)

# Download requested but not yet begun.
PENDING = 'pending'
# Actively downloading.
DOWNLOADING = 'downloading'
# Download complete, not necessarily finished callback.
DOWNLOADED = 'downloaded'
# Downloaded and finished processing callback, if any.
COMPLETE = 'complete'
# Exception occurred -- could be during download or callback.
EXCEPTION = 'exception'
# Aborted
ABORTED = 'aborted'
ALL_STATES = set([PENDING, DOWNLOADING, DOWNLOADED, COMPLETE, EXCEPTION, ABORTED])
ACTIVE_STATES = set([PENDING, DOWNLOADING, DOWNLOADED])
TERMINAL_STATES = ALL_STATES - ACTIVE_STATES

TEMP_FILE_SUFFIX = '.part'

# Keep in sync with master_agent.avdl
URL_FIELD = 'url'
STATE_FIELD = 'state'
ACTIVE_FIELD = 'active'
PROGRESS_FIELD = 'progress'
ERROR_FIELD = 'parcel_err'
ERROR_CODE_FIELD = 'code'
ERROR_MSG_FIELD = 'msg'

UNEXPECTED_ERROR_CODE = 600
CALLBACK_ERROR_CODE = 601

class URLopener(urllib.FancyURLopener):
  """
  Subclassed opener so we can throw exceptions on http errors.
  """
  def http_error_default(self, url, fp, code, msg, headers):
    fp.close()
    raise urllib2.HTTPError(url, code, msg, headers, fp)

  def rename(self, src, dest):
    '''This is here just to simplify mocking for tests'''
    os.rename(src, dest)

class Download(object):

  def __init__(self, url):
    self.url = url
    self._state = PENDING
    self.progress = -1
    self.code = 0
    self.exception = None
    self.path = None
    self.headers = None

  def get_state(self):
    return self._state

  def set_state(self, state):
    if self._state in TERMINAL_STATES:
      return
    else:
      self._state = state

  state = property(get_state, set_state)

  def __repr__(self):
    return "[ url: %s, state: %s, progress: %d, code: %d, exception: %s, path: %s ]" % (self.url, self.state, self.progress, self.code, str(self.exception), self.path)

class Downloader(object):
  """
  Class that implements our parcel downloader.
  """

  def __init__(self, path, limit, opener=URLopener()):
    LOG.info("Downloader path: %s" % (path))
    self.path = path
    self.limit = limit
    self.threadpool = threadpool.ThreadPool(limit)
    self.ops = dict()
    self.ops_lock = threading.Lock()
    self.opener = opener

  def cleanUp(self):
    self.threadpool.dismissWorkers(self.limit)

  def wait(self):
    return self.threadpool.wait()

  def poll(self):
    return self.threadpool.poll()

  def queue_download(self, url, callback=None, overwrite=False, proceed=False):
    """Queue a download for the url.  If there is already a download for the
       url in active state, the request will be ignored with False returned. If
       there is a download in terminal state, it will be replaced, unless
       'proceed' is True and the state is complete. This is intended for
       retrying callback with successfully downloaded file. The 'overwrite'
       flag can be used to first delete any existing downloaded file. True is
       returned if the request is enqueued.

       The optional callback function takes as argument the url and a Download
       object that contains information about the download. The function will
       be called regardless of the outcome of the download, and should return
       True if completes, causing the state to transition to COMPLETE. If no
       callback and download is DOWNLOADED, the state would also transition to
       COMPLETE.
    """
    downloaded_path = self._get_downloaded_path(url)
    already_exist = os.path.exists(downloaded_path)
    if already_exist and overwrite:
      LOG.info("Removing existing file %s" % (downloaded_path))
      os.remove(downloaded_path)
      already_exist = False

    try:
      self.ops_lock.acquire()

      if already_exist and url not in self.ops:
        # fake a complete op (ops state may be lost due to agent restart)
        op = Download(url)
        op.state = DOWNLOADED
        op.code = 200
        op.path = downloaded_path
        self.ops[url] = op

      if url in self.ops:
        curr_op = self.ops[url]
        if curr_op.state is DOWNLOADED and proceed:
          LOG.debug("Proceeding with callback for downloaded download %s" % (url))
        elif curr_op.state in ACTIVE_STATES:
          LOG.info("There is already an active download for %s" % (url))
          return False
        elif curr_op.state in TERMINAL_STATES:
          # overwrite existing op
          curr_op = Download(url)
        else:
          raise Exception("Unexpected state %s" % (op.state))
      else:
        curr_op = Download(url)
      self.ops[url] = curr_op
    finally:
      self.ops_lock.release()

    def callable():
      LOG.debug("Executing request for op %s" % (str(curr_op)))
      if curr_op.state is PENDING: # Could be DOWNLOADED for an existing file.
        self._download(curr_op)
      # directly process our callback using the threadpool's threads. we choose
      # this over threadpool's callback mechanism because the latter relies on
      # calling poll() with our own thread, and callback can potentially take
      # a long time (e.g. untar huge parcel).
      if callback is None:
        if curr_op.state is DOWNLOADED:
          curr_op.state = COMPLETE
      else:
        try:
          if callback(url, curr_op) and curr_op.state is DOWNLOADED:
            curr_op.state = COMPLETE
        except Exception, e:
          self._set_exception(curr_op, CALLBACK_ERROR_CODE, e)
          self._remove(curr_op.path)
      return curr_op

    request = threadpool.WorkRequest(callable)
    self.threadpool.putRequest(request)
    LOG.debug("Queued download for %s into %s" % (url, downloaded_path))
    return True

  def abort(self, url):
    try:
      self.ops_lock.acquire()
      if url not in self.ops:
        LOG.info("No download for url %s" % (url))
        return False
      op = self.ops[url]
      if op.state in ACTIVE_STATES:
        LOG.info("Aborting download request for %s" % (url))
        op.state = ABORTED
        # TODO : Abort job in pool, when we switch to a library that allows it.
        return True
      else:
        LOG.debug("Ignoring abort for %s because its state %s is terminal" % (url, op.state))
        return False
    finally:
      self.ops_lock.release()

  def prune_history(self, filter):
    """Prune download history of urls that satisfy the provided filter
       function. The function will be invoked with one argument the
       url string, and should return True if it is to be pruned.
    """
    try:
      self.ops_lock.acquire()
      pruned = dict()
      for op in self.ops.itervalues():
        if not filter(op.url):
          pruned[op.url] = op
      self.ops = pruned
    finally:
      self.ops_lock.release()

  def _get_downloaded_path(self, url):
    o = urlparse.urlparse(url)
    # o[2] should ideally be o.path, but in 2.4 it's just a tuple
    return os.path.join(self.path, os.path.basename(o[2]))

  def _download(self, op):
    '''Download file from any url'''
    url = op.url

    filename = self._get_downloaded_path(url)
    tempname = filename + TEMP_FILE_SUFFIX
    self._remove(tempname)

    def cb(*args, **kwargs):
      kwargs['url'] = url
      self.progress(*args, **kwargs)

    op.state = DOWNLOADING
    LOG.info("Starting download of: %s", url)

    try:
      dlname, op.headers = self.opener.retrieve(url, tempname, cb)
    except urllib2.HTTPError, e:
      LOG.exception("HTTP error during download")
      self._set_exception(op, e.code, e)
      self._remove(tempname)
      return
    except Exception, e:
      LOG.exception("Unexpected exception during download")
      self._set_exception(op, UNEXPECTED_ERROR_CODE, e)
      self._remove(tempname)
      return

    op.state = DOWNLOADED
    op.code = 200

    try:
      self.opener.rename(tempname, filename)
      op.path = filename
    except Exception, e:
      op.state = EXCEPTION
      op.exception = e

    LOG.info("Completed download of %s code=%d state=%s", url, op.code, str(op.state))
    return op

  def _remove(self, file):
    if os.path.exists(file):
      try:
        os.remove(file)
      except OSError, e:
        LOG.exception("Error removing file %s", (file))
        raise e

  def _set_exception(self, op, code, exception):
      op.state = EXCEPTION
      op.exception = exception
      op.code = code

  def progress(self, block_count, block_size, total_size, url):
    if total_size == -1:
      self.ops[url].progress = -1
    else:
      self.ops[url].progress = (100 * block_count * block_size)/total_size

  def get_current_downloads(self):
    dws = []
    for op in self.ops.values():
      dw = dict()
      dw[URL_FIELD] = op.url
      dw[STATE_FIELD] = op.state
      dw[ACTIVE_FIELD] = op.state in ACTIVE_STATES
      dw[PROGRESS_FIELD] = op.progress
      if (op.state == EXCEPTION):
        err = dict()
        err[ERROR_CODE_FIELD] = str(op.code)
        err[ERROR_MSG_FIELD] = str(op.exception)
        dw[ERROR_FIELD] = err
      else:
        dw[ERROR_FIELD] = None
      dws.append(dw)
    return dws

if __name__ == '__main__':
  if len(sys.argv) < 4:
    print "downloader.py <download dir> <max simultaneous downloads> [urls...]"
    sys.exit(1)

  LOG_FORMAT = '[%(asctime)s] %(process)d %(threadName)s %(module)-12s %(levelname)-8s %(message)s'
  DATE_FORMAT = '%d/%b/%Y %H:%M:%S %z'
  logging.basicConfig(level=logging.DEBUG,
    format=LOG_FORMAT, datefmt=DATE_FORMAT)

  downloader = Downloader(sys.argv[1], int(sys.argv[2]))

  def callback(url):
    print downloader.ops[url]

  for url in sys.argv[3:]:
    downloader.queue_download(url, callback=callback)

  try:
    while True:
      for url, ops in downloader.ops.items():
        print "%s:%d" % (url, ops.progress)
      downloader.poll()
  except:
    # downloader.poll() throws an exception when there's no work left.
    pass

  downloader.cleanUp()
