// Copyright (c) 2012 Cloudera, Inc.  All rights reserved.
define([
  "knockout",
  "underscore",
  "cloudera/Util",
  "cloudera/common/UserSettings",
  "cloudera/cmf/wizard/WizardBase",
  "cloudera/cmf/wizard/hosts/RepositorySelectionStep",
  "cloudera/cmf/wizard/hosts/HostCredentialsStep",
  "cloudera/cmf/wizard/hosts/InstallStep",
  "cloudera/cmf/wizard/hosts/HostInspectorStep",
  "cloudera/cmf/wizard/express/ClusterCreationStep",
  "cloudera/cmf/wizard/express/ParcelInstallStep"
], function (ko, _, Util, UserSettings, WizardBase, RepositorySelectionStep, HostCredentialsStep, InstallStep, HostInspectorStep, ClusterCreationStep, ParcelInstallStep) {
  "use strict";

  /**
   * options = {
   *   hostInspectorJsonUrl: (required) "URL to get the JSON structure that describes the running of a hostInspector",
   *   parcelUpdateUrl:      (required) "URL to get the JSON structure that describes the latest parcel status",
   *   parcelInstallUrl:     (required) "URL to install parcel",
   *   parcelListUrl:        (required) "URL to list parcels",
   *   parcelAddRepoUrl:     (required) "URL to add parcel repo",
   *   container:            (required) "parameter required for WizardBase",
   *   exitUrl:              (required) "parameter required for WizardBase",
   *   settingsKey:          (required) "the settings key to write to",
   *   existingHosts:        (optional) ["hostNames"],
   *   newHosts:             (optional) ["hostNames"],
   *   resumeState:          (optional) the state where we should resume the wizard to.
   * }
   */
  function ExpressAddHostsWizard(options) {
    var self = this, steps = [];

    var resumeState = options.resumeState || {};
    var firstTry = options.resumeState === undefined;

    var getHosts = function() {
      if (resumeState.hosts) {
        return resumeState.hosts;
      } else {
        if (_.isEmpty(options.existingHosts)) {
          return options.newHosts || [];
        } else {
          return options.existingHosts || [];
        }
      }
    };

    // each id must match the id of the parent container.
    // We are passing the id here because
    // it is also used in the URL hash.

    // Always applicable
    var repositorySelectionStep = new RepositorySelectionStep({
      id: "repositorySelectionStep",
      listUrl: options.parcelListUrl,
      addCustomRepoUrl: options.parcelAddRepoUrl,
      enableFormatSelection: true,
      enableParcelSelection: true,
      enablePackageSelection: !_.isEmpty(options.newHosts),
      // Ask for version regardless of new/existing hosts. RepositorySelectionStep only shows if "usePackage".
      enableCDHVersionSelection: true,
      enableCM: !_.isEmpty(options.newHosts)
    });
    if (firstTry) {
      steps.push(repositorySelectionStep);
    }

    var getCdhVersion = function() {
      if (resumeState.cdhVersion !== undefined) {
        return resumeState.cdhVersion;
      } else {
        if (repositorySelectionStep.installMethod() === "useParcel") {
          return repositorySelectionStep.cdhVersionInParcel();
        } else {
          return repositorySelectionStep.cdhVersion();
        }
      }
    };

    var getInstallMethod = function() {
      if (resumeState.installMethod !== undefined) {
        return resumeState.installMethod;
      } else {
        return repositorySelectionStep.installMethod();
      }
    };

    var getParcelsFunction = function() {
      if (resumeState.parcels !== undefined) {
        return function() {
          return resumeState.parcels;
        };
      } else {
        return repositorySelectionStep.chosenParcels;
      }
    };

    var getHasNewHosts = function() {
      if (resumeState.hasNewHosts !== undefined) {
        return resumeState.hasNewHosts;
      } else {
        return !_.isEmpty(options.newHosts);
      }
    };

    // Skip unless there are newHosts.
    var installStep = new InstallStep({
      id: "installStep",
      hostInstallUrl: resumeState.hostInstallUrl
    });

    if (getHasNewHosts()) {
      if (firstTry) {
        steps.push(new HostCredentialsStep({
          id: "hostCredentialsStep"
        }));
      }

      if (!resumeState.installedHosts) {
        steps.push(installStep);
      }
    }

    // Create cluster with either newly installed or existing hosts.
    var clusterCreationStep = new ClusterCreationStep({
      id: "clusterCreationStep",
      createUrl: "/cmf/express-wizard/create-cluster",
      deleteUrl: "/cmf/express-wizard/delete-cluster",
      hosts: getHosts,
      version: getCdhVersion
    });
    if (!resumeState.createdCluster) {
      steps.push(clusterCreationStep);
    }
    var getClusterFunction = function() {
      if (resumeState.cluster !== undefined) {
        return function() {
          return resumeState.cluster;
        };
      } else {
        return clusterCreationStep.cluster;
      }
    };

    var parcelInstallStep = new ParcelInstallStep({
      id: "parcelInstallStep",
      updateUrl: options.parcelUpdateUrl,
      installUrl: options.parcelInstallUrl,
      parcels: getParcelsFunction(),
      cluster: getClusterFunction(),
      shouldSkip: function() {
        return getInstallMethod() === "usePackage";
      },
      displayPageWhenActivated: true
    });

    // Skip if usePackage
    if (!resumeState.installedParcels) {
      steps.push(parcelInstallStep);
    }

    var hostInspectorStep = new HostInspectorStep({
      id: "hostInspectorStep",
      url: options.hostInspectorJsonUrl
    });

    if (!resumeState.inspectedHosts) {
      steps.push(hostInspectorStep);
    }

    var getCurrentState = function() {
      return {
        hasNewHosts: getHasNewHosts(),
        hostInstallUrl: installStep.getHostInstallUrl(),
        installedHosts: resumeState.installedHosts,
        inspectedHosts: resumeState.inspectedHosts,
        createdCluster: resumeState.createdCluster,
        installedParcels: resumeState.installedParcels,
        parcels: getParcelsFunction()(),
        cluster: getClusterFunction()(),
        hosts: getHosts(),
        installMethod: getInstallMethod(),
        cdhVersion: getCdhVersion()
      };
    };

    this.wizard = new WizardBase({
      container: options.container,
      exitUrl: options.exitUrl,
      steps: steps
    });

    var persistState = function() {
      // Save this to the backend.
      var currentState = getCurrentState();
      UserSettings.update(options.settingsKey, JSON.stringify(currentState));
    };

    this.wizard.onEnterStep = function(value) {
      if (value > 0) {
        var previousStep = self.wizard.steps[value - 1];
        var persist = false;
        if (previousStep === installStep) {
          resumeState.installedHosts = true;
          persist = true;
        } else if (previousStep === clusterCreationStep) {
          resumeState.createdCluster = true;
          persist = true;
        } else if (previousStep === parcelInstallStep) {
          resumeState.installedParcels = true;
          persist = true;
        } else if (previousStep === hostInspectorStep) {
          resumeState.inspectedHosts = true;
          persist = true;
        }
        // We only want to persist when we are leaving the selected step.
        if (persist) {
          persistState();
        }
      }
    };

    this.wizard.leave = function() {
      UserSettings.update(options.settingsKey, "", function() {
        Util.setWindowLocation(options.exitUrl + "?clusterId=" + getClusterFunction()().id);
      });
    };

    var handle1 = $.subscribe("hostInstallUrlChanged", function(hostInstallUrl) {
      resumeState.hostInstallUrl = hostInstallUrl;
      persistState();
    });
    this.subscriptionHandles = [handle1];
    this.unsubscribe = function() {
      Util.unsubscribe(this.wizard);
      Util.unsubscribe(this);
    };
  }
  return ExpressAddHostsWizard;
});
